"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MiniSignal = void 0;
const mini_signals_binding_1 = require("./mini-signals-binding");
class MiniSignal {
    constructor() {
        this._head = undefined;
        this._tail = undefined;
    }
    hasHandlers() {
        return !(this._head == null);
    }
    /**
     * Return an array of attached MiniSignalBinding.
     */
    handlers() {
        let node = this._head;
        const ee = [];
        while (node != null) {
            ee.push(node);
            node = node._next;
        }
        return ee;
    }
    /**
     * Return true if node is a MiniSignalBinding attached to this MiniSignal
     */
    has(node) {
        if (!(node instanceof mini_signals_binding_1.MiniSignalBinding)) {
            throw new Error('MiniSignal#has(): First arg must be a MiniSignalBinding object.');
        }
        return node._owner === this;
    }
    /**
     * Dispatches a signal to all registered listeners.
     */
    dispatch(...args) {
        let node = this._head;
        if (node == null)
            return false;
        while (node != null) {
            if (node._once)
                this.detach(node);
            node._fn.apply(node._thisArg, args);
            node = node._next;
        }
        return true;
    }
    /**
     * Register a new listener.
     */
    add(fn, thisArg) {
        if (typeof fn !== 'function') {
            throw new Error('MiniSignal#add(): First arg must be a Function.');
        }
        return this._addMiniSignalBinding(new mini_signals_binding_1.MiniSignalBinding(fn, false, thisArg));
    }
    /**
     * Register a new listener that will be executed only once.
     */
    once(fn, thisArg) {
        if (typeof fn !== 'function') {
            throw new Error('MiniSignal#once(): First arg must be a Function.');
        }
        return this._addMiniSignalBinding(new mini_signals_binding_1.MiniSignalBinding(fn, true, thisArg));
    }
    /**
     * Remove binding object.
     */
    detach(node) {
        if (!(node instanceof mini_signals_binding_1.MiniSignalBinding)) {
            throw new Error('MiniSignal#detach(): First arg must be a MiniSignalBinding object.');
        }
        if (node._owner !== this)
            return this; // todo: or error?
        if (node._prev !== undefined && node._prev !== null)
            node._prev._next = node._next;
        if (node._next !== undefined && node._next !== null)
            node._next._prev = node._prev;
        if (node === this._head) {
            // first node
            this._head = node._next;
            if (node._next === null) {
                this._tail = undefined;
            }
        }
        else if (node === this._tail) {
            // last node
            this._tail = node._prev;
            // eslint-disable-next-line  @typescript-eslint/no-non-null-assertion
            this._tail._next = undefined;
        }
        node._owner = null;
        return this;
    }
    /**
     * Detach all listeners.
     */
    detachAll() {
        let node = this._head;
        if (node == null)
            return this;
        this._head = this._tail = undefined;
        while (node != null) {
            node._owner = null;
            node = node._next;
        }
        return this;
    }
    _addMiniSignalBinding(node) {
        if (this._head == null) {
            this._head = node;
            this._tail = node;
        }
        else {
            // eslint-disable-next-line  @typescript-eslint/no-non-null-assertion
            this._tail._next = node;
            node._prev = this._tail;
            this._tail = node;
        }
        node._owner = this;
        return node;
    }
}
exports.MiniSignal = MiniSignal;
