"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const mini_signals_1 = require("../src/mini-signals");
const mini_signals_binding_1 = require("../src/mini-signals-binding");
const node_util_1 = require("node:util");
const chai_1 = require("chai");
const node_assert_1 = require("node:assert");
describe('MiniSignal', () => {
    it('inherits when used with require(util).inherits', () => {
        function Beast() {
            /* rawr, i'm a beast */
        }
        (0, node_util_1.inherits)(Beast, mini_signals_1.MiniSignal);
        const moop = new Beast();
        const meap = new Beast();
        (0, chai_1.expect)(moop).is.instanceOf(Beast);
        (0, chai_1.expect)(moop).is.instanceOf(mini_signals_1.MiniSignal);
        moop.handlers();
        meap.handlers();
        /* istanbul ignore next */
        moop.add(() => {
            throw new Error('I should not dispatch');
        });
        meap.dispatch('rawr');
        meap.detachAll();
    });
    it('quick test', () => {
        const pattern = [];
        const e = new mini_signals_1.MiniSignal();
        const foo = e.add(writer, 'foo');
        e.add(writer, 'baz');
        const bar = e.add(writer, 'bar');
        (0, chai_1.expect)(e instanceof mini_signals_1.MiniSignal);
        (0, chai_1.expect)(foo instanceof mini_signals_binding_1.MiniSignalBinding);
        e.dispatch('banana');
        e.dispatch('apple');
        foo.detach();
        bar.detach();
        e.dispatch('pear');
        e.detachAll();
        e.dispatch('raspberry');
        (0, chai_1.expect)(pattern.join(';') ===
            'foo:banana;baz:banana;bar:banana;foo:apple;baz:apple;bar:apple;baz:pear');
        function writer(a) {
            pattern.push(String(this) + ':' + a);
        }
    });
    describe('MiniSignal#once', () => {
        let e;
        let context;
        beforeEach(() => {
            e = new mini_signals_1.MiniSignal();
            context = { bar: 'baz' };
        });
        it('should throw error for incorrect types', () => {
            (0, chai_1.expect)(() => {
                // @ts-expect-error testing error
                e.once();
            }).throws('MiniSignal#once(): First arg must be a Function.');
            (0, chai_1.expect)(() => {
                // @ts-expect-error testing error
                e.once(123);
            }).throws('MiniSignal#once(): First arg must be a Function.');
            (0, chai_1.expect)(() => {
                // @ts-expect-error testing error
                e.once(true);
            }).throws('MiniSignal#once(): First arg must be a Function.');
            (0, chai_1.expect)(e.handlers().length).equals(0);
        });
        it('should not invoke twice', () => {
            const cb = function (bar) {
                (0, chai_1.expect)(bar).equals('bar');
                (0, chai_1.expect)(this).equals(context);
                (0, chai_1.expect)(arguments).has.length(1);
                e.dispatch('bar');
            }.bind(context);
            e.once(cb);
            e.dispatch('bar');
        });
    });
    describe('MiniSignal#add', () => {
        let e;
        beforeEach(() => {
            e = new mini_signals_1.MiniSignal();
        });
        it('should throw error for incorrect types', () => {
            (0, chai_1.expect)(() => {
                // @ts-expect-error testing error
                e.add();
            }).throws('MiniSignal#add(): First arg must be a Function.');
            (0, chai_1.expect)(() => {
                // @ts-expect-error testing error
                e.add(123);
            }).throws('MiniSignal#add(): First arg must be a Function.');
            (0, chai_1.expect)(() => {
                // @ts-expect-error testing error
                e.add(true);
            }).throws('MiniSignal#add(): First arg must be a Function.');
            (0, chai_1.expect)(e.handlers().length).equals(0);
        });
    });
    describe('MiniSignal#dispatch', () => {
        function writer() {
            pattern += String(this);
        }
        let e;
        let context = { bar: 'baz' };
        let pattern = '';
        beforeEach(() => {
            e = new mini_signals_1.MiniSignal();
            context = { bar: 'baz' };
            pattern = '';
        });
        it('should return false when there are not events to dispatch', () => {
            (0, chai_1.expect)(e.dispatch('foo')).equals(false);
            (0, chai_1.expect)(e.dispatch('bar')).equals(false);
        });
        it('emits with context when function is bound function', () => {
            const cb = function (bar) {
                (0, chai_1.expect)(bar).equals('bar');
                (0, chai_1.expect)(this).equals(context);
                (0, chai_1.expect)(arguments).has.length(1);
            }.bind(context);
            e.add(cb);
            e.dispatch('bar');
        });
        it('emits with context when context is specified', () => {
            e.add(function (bar) {
                (0, chai_1.expect)(bar).equals('bar');
                (0, chai_1.expect)(this).equals(context);
                (0, chai_1.expect)(arguments).has.length(1);
            }, context);
            e.dispatch('bar');
        });
        it('can dispatch the function with multiple arguments', () => {
            for (let i = 0; i < 100; i++) {
                const e = new mini_signals_1.MiniSignal();
                (function (j) {
                    const args = [];
                    for (let i = 0; i < j; i++) {
                        args.push(j);
                    }
                    e.add((..._args) => {
                        (0, chai_1.expect)(_args.length).equals(args.length);
                    });
                    e.dispatch.apply(e, args);
                })(i);
            }
        });
        it('can dispatch the function with multiple arguments, multiple listeners', () => {
            for (let i = 0; i < 100; i++) {
                const e = new mini_signals_1.MiniSignal();
                (function (j) {
                    const args = [];
                    for (let i = 0; i < j; i++) {
                        args.push(j);
                    }
                    e.add((..._args) => {
                        (0, chai_1.expect)(_args.length).equals(args.length);
                    });
                    e.add((..._args) => {
                        (0, chai_1.expect)(_args.length).equals(args.length);
                    });
                    e.add((..._args) => {
                        (0, chai_1.expect)(_args.length).equals(args.length);
                    });
                    e.add((..._args) => {
                        (0, chai_1.expect)(_args.length).equals(args.length);
                    });
                    e.dispatch.apply(e, args);
                })(i);
            }
        });
        it('can dispatch many listeners', () => {
            const N = 10000;
            let sum = 0;
            function add(i) {
                sum += i;
            }
            for (let i = 0; i <= N; i++) {
                e.add(add.bind(this, i));
            }
            e.dispatch();
            (0, chai_1.expect)(sum).equals((N * (N + 1)) / 2);
        });
        it('emits with context, multiple listeners (force loop)', () => {
            e.add(function (bar) {
                (0, chai_1.expect)(this).deep.equals({ foo: 'bar' });
                (0, chai_1.expect)(bar).equals('bar');
            }, { foo: 'bar' });
            e.add(function (bar) {
                (0, chai_1.expect)(this).deep.equals({ bar: 'baz' });
                (0, chai_1.expect)(bar).equals('bar');
            }, { bar: 'baz' });
            e.dispatch('bar');
        });
        it('emits with different contexts', () => {
            e.add(writer, 'foo');
            e.add(writer, 'baz');
            e.add(writer, 'bar');
            e.add(writer, 'banana');
            e.dispatch();
            (0, chai_1.expect)(pattern).equals('foobazbarbanana');
        });
        it('should return true when there are events to dispatch', function (done) {
            e.add(() => {
                process.nextTick(done);
            });
            (0, chai_1.expect)(e.dispatch()).equals(true);
        });
        it('should return false when there are no events to dispatch', () => {
            (0, chai_1.expect)(e.dispatch()).equals(false);
        });
        it('receives the emitted events', (done) => {
            const e = new mini_signals_1.MiniSignal();
            e.add(function (a, b, c, d, undef) {
                (0, chai_1.expect)(a).equals('foo');
                (0, chai_1.expect)(b).equals(e);
                (0, chai_1.expect)(c).is.instanceOf(Date);
                (0, chai_1.expect)(undef).equals(undefined);
                (0, chai_1.expect)(arguments).has.length(3);
                done();
            });
            e.dispatch('foo', e, new Date());
        });
        it('emits to all event listeners', () => {
            const e = new mini_signals_1.MiniSignal();
            const pattern = [];
            e.add(() => {
                pattern.push('foo1');
            });
            e.add(() => {
                pattern.push('foo2');
            });
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('foo1;foo2');
        });
        it('emits to all event listeners', () => {
            const e = new mini_signals_1.MiniSignal();
            const pattern = [];
            function foo1() {
                pattern.push('foo1');
            }
            function foo2() {
                pattern.push('foo2');
            }
            function foo3() {
                pattern.push('foo3');
            }
            e.add(foo1);
            e.add(foo2);
            e.add(foo3);
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('foo1;foo2;foo3');
        });
        it('emits to all event listeners, removes once', () => {
            const e = new mini_signals_1.MiniSignal();
            const pattern = [];
            function foo1() {
                pattern.push('foo1');
            }
            function foo2() {
                pattern.push('foo2');
            }
            function foo3() {
                pattern.push('foo3');
            }
            e.add(foo1);
            e.once(foo2);
            e.add(foo3);
            e.dispatch();
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('foo1;foo2;foo3;foo1;foo3');
        });
    });
    describe('MiniSignal#handlers', () => {
        /* istanbul ignore next */
        function foo() { }
        it('returns an empty array if no handlers are added', () => {
            const e = new mini_signals_1.MiniSignal();
            (0, chai_1.expect)(e.handlers()).is.a('array');
            (0, chai_1.expect)(e.handlers().length).equals(0);
        });
        it('returns an array of MiniSignalBinding', () => {
            const e = new mini_signals_1.MiniSignal();
            e.add(foo);
            e.add(foo);
            (0, chai_1.expect)(e.handlers()).is.a('array');
            (0, chai_1.expect)(e.handlers().length).equals(2);
            e.handlers().forEach(function (h) {
                (0, chai_1.expect)(h).instanceOf(mini_signals_binding_1.MiniSignalBinding);
            });
        });
        it('is not vulnerable to modifications', () => {
            const e = new mini_signals_1.MiniSignal();
            e.add(foo);
            e.add(foo);
            (0, chai_1.expect)(e.handlers().length).equals(2);
            e.handlers().length = 0;
            (0, chai_1.expect)(e.handlers().length).equals(2);
            e.handlers().forEach(function (h) {
                (0, chai_1.expect)(h).instanceOf(mini_signals_binding_1.MiniSignalBinding);
            });
        });
        it('can return a boolean as indication if handlers exist', () => {
            const e = new mini_signals_1.MiniSignal();
            e.add(foo);
            e.add(foo);
            e.add(foo);
            e.add(foo);
            e.add(foo);
            e.add(foo);
            (0, chai_1.expect)(e.hasHandlers()).equals(true);
            e.detachAll();
            (0, chai_1.expect)(e.hasHandlers()).equals(false);
        });
    });
    describe('MiniSignal#detach', () => {
        /* istanbul ignore next */
        function foo() {
            pattern.push('foo');
        }
        /* istanbul ignore next */
        function bar() {
            pattern.push('bar');
        }
        /* istanbul ignore next */
        function a() {
            pattern.push('a');
        }
        /* istanbul ignore next */
        function b() {
            pattern.push('b');
        }
        let e;
        let pattern = [];
        beforeEach(() => {
            e = new mini_signals_1.MiniSignal();
            pattern = [];
        });
        it('should throw an error if not a SignalBinding', () => {
            (0, chai_1.expect)(() => {
                // @ts-expect-error testing error
                e.detach();
            }).throws('MiniSignal#detach(): First arg must be a MiniSignalBinding object.');
            (0, chai_1.expect)(() => {
                // @ts-expect-error testing error
                e.detach(1);
            }).throws('MiniSignal#detach(): First arg must be a MiniSignalBinding object.');
            (0, chai_1.expect)(() => {
                // @ts-expect-error testing error
                e.detach(bar);
            }).throws('MiniSignal#detach(): First arg must be a MiniSignalBinding object.');
        });
        it('should only remove the event with the specified node', () => {
            e.add(a);
            e.add(b);
            const _bar = e.add(bar);
            (0, chai_1.expect)(e.handlers().length).equals(3);
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('a;b;bar');
            e.detach(_bar);
            (0, chai_1.expect)(e.handlers().length).equals(2);
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('a;b;bar;a;b');
        });
        it('should remove from front', () => {
            const _bar = e.add(bar);
            e.add(a);
            e.add(b);
            (0, chai_1.expect)(e.handlers().length).equals(3);
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('bar;a;b');
            e.detach(_bar);
            (0, chai_1.expect)(e.handlers().length).equals(2);
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('bar;a;b;a;b');
        });
        it('should remove from middle', () => {
            e.add(a);
            const _bar = e.add(bar);
            e.add(b);
            (0, chai_1.expect)(e.handlers().length).equals(3);
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('a;bar;b');
            e.detach(_bar);
            (0, chai_1.expect)(e.handlers().length).equals(2);
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('a;bar;b;a;b');
        });
        it('emits to all event listeners after removing', () => {
            e.add(a);
            const _foo = e.add(foo);
            e.add(b);
            e.detach(_foo);
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('a;b');
        });
        it('can remove previous node in dispatch', () => {
            const _foo = e.add(foo);
            e.add(foo2);
            e.add(a);
            e.dispatch();
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('foo;foo2;a;foo2;a');
            function foo2() {
                pattern.push('foo2');
                e.detach(_foo);
            }
        });
        it('can remove next node in dispatch', () => {
            e.add(a);
            e.add(foo2);
            const _foo = e.add(foo);
            e.dispatch();
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('a;foo2;a;foo2'); // will remove node this dispatch (might be unexpected)
            function foo2() {
                pattern.push('foo2');
                e.detach(_foo);
            }
        });
        it('can remove node in dispatch', () => {
            e.add(foo2);
            e.add(a);
            const _foo = e.add(foo);
            e.dispatch();
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('foo2;a;foo2;a'); // will remove node this dispatch (might be unexpected)
            function foo2() {
                pattern.push('foo2');
                e.detach(_foo);
            }
        });
        it('can remove current node in dispatch', () => {
            e.add(a);
            const _foo = e.add(foo2);
            e.add(b);
            e.dispatch();
            e.dispatch();
            (0, chai_1.expect)(pattern.join(';')).equals('a;foo2;b;a;b');
            function foo2() {
                pattern.push('foo2');
                e.detach(_foo);
            }
        });
        it('can only detach from same signal', () => {
            const e2 = new mini_signals_1.MiniSignal();
            const binding = e.add(foo);
            e2.detach(binding);
            (0, chai_1.expect)(binding._owner === e);
            (0, chai_1.expect)(e.hasHandlers());
        });
        it('can be called multiple times', () => {
            const binding = e.add(foo);
            (0, chai_1.expect)(binding._owner === e);
            e.detach(binding);
            (0, chai_1.expect)(binding._owner === null);
            e.detach(binding);
            e.detach(binding);
        });
    });
    describe('MiniSignal#detachAll', () => {
        /* istanbul ignore next */
        function oops() {
            throw new Error('oops');
        }
        let e;
        beforeEach(() => {
            e = new mini_signals_1.MiniSignal();
        });
        it('removes all events', () => {
            e.add(oops);
            e.add(oops);
            e.add(oops);
            e.add(oops);
            (0, chai_1.expect)(e.handlers().length).equals(4);
            (0, chai_1.expect)(e.detachAll()).equals(e);
            (0, chai_1.expect)(e.handlers().length).equals(0);
            (0, chai_1.expect)(e.dispatch()).equals(false);
        });
        it('should not throw an error if no listeners are set', () => {
            (0, chai_1.expect)(e.detachAll()).equals(e);
            (0, chai_1.expect)(e.handlers().length).equals(0);
            (0, chai_1.expect)(e.dispatch()).equals(false);
        });
        it('Should not throw error when calling detach after detachAll', () => {
            const binding = e.add(oops);
            e.detachAll();
            e.detach(binding);
        });
    });
    describe('MiniSignal#has', () => {
        /* istanbul ignore next */
        function oops() {
            throw new Error('oops');
        }
        let e;
        beforeEach(() => {
            e = new mini_signals_1.MiniSignal();
        });
        it('has returns true if bound', () => {
            const binding = e.add(oops);
            (0, node_assert_1.strict)(e.has(binding));
        });
        it('has returns false if bound to another signal', () => {
            const e2 = new mini_signals_1.MiniSignal();
            const binding = e2.add(oops);
            (0, node_assert_1.strict)(e.has(binding) === false);
        });
        it('has returns false if detached', () => {
            const binding = e.add(oops);
            (0, node_assert_1.strict)(e.has(binding));
            binding.detach();
            (0, node_assert_1.strict)(e.has(binding) === false);
        });
        it('has returns false after detachAll', () => {
            const binding = e.add(oops);
            (0, node_assert_1.strict)(e.has(binding));
            e.detachAll();
            (0, node_assert_1.strict)(e.has(binding) === false);
        });
        it('should throw error for incorrect types', () => {
            (0, chai_1.expect)(() => {
                e.has({});
            }).throws('MiniSignal#has(): First arg must be a MiniSignalBinding object.');
        });
    });
    describe('Readme Examples', () => {
        it('Example Usage', () => {
            const mySignal = new mini_signals_1.MiniSignal();
            const binding = mySignal.add(onSignal); // add listener
            mySignal.dispatch('foo', 'bar'); // dispatch signal passing custom parameters
            binding.detach(); // remove a single listener
            function onSignal(foo, bar) {
                (0, node_assert_1.strict)(foo === 'foo');
                (0, node_assert_1.strict)(bar === 'bar');
            }
        });
        it('Another Example', () => {
            const myObject = {
                foo: 'bar',
                updated: new mini_signals_1.MiniSignal(),
            };
            myObject.updated.add(onUpdated, myObject); // add listener with context
            myObject.foo = 'baz';
            myObject.updated.dispatch(); // dispatch signal
            function onUpdated() {
                (0, node_assert_1.strict)(this === myObject);
                (0, node_assert_1.strict)(this.foo === 'baz');
            }
        });
        it('Function#bind example', () => {
            const mySignal = new mini_signals_1.MiniSignal();
            const context = {};
            const cb = function (bar) {
                (0, chai_1.expect)(arguments).has.length(1);
                (0, chai_1.expect)(bar).equals('bar');
                (0, chai_1.expect)(this).equals(context);
            }.bind(context);
            mySignal.add(cb);
            mySignal.dispatch('bar');
        });
        it('Function#bind example with parameters', () => {
            const mySignal = new mini_signals_1.MiniSignal();
            const context = {};
            const cb = function (bar) {
                (0, chai_1.expect)(arguments).has.length(1);
                (0, chai_1.expect)(bar).equals('bar');
                (0, chai_1.expect)(this).equals(context);
            }.bind(context, 'bar');
            mySignal.add(cb);
            mySignal.dispatch();
        });
    });
});
